/*
 * Copyright (c) 2002-2016 "Neo Technology,"
 * Network Engine for Objects in Lund AB [http://neotechnology.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.neo4j.cypher.internal.javacompat;

import org.hamcrest.Matcher;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import org.neo4j.cypher.javacompat.internal.GraphDatabaseCypherService;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.Transaction;
import org.neo4j.kernel.api.AccessMode;
import org.neo4j.kernel.api.KernelTransaction;
import org.neo4j.kernel.impl.query.QueryEngineProvider;
import org.neo4j.kernel.impl.query.QuerySession;
import org.neo4j.logging.NullLogProvider;
import org.neo4j.test.TestGraphDatabaseFactory;

import static org.hamcrest.Matchers.not;
import static org.junit.Assert.assertThat;
import static org.neo4j.graphdb.Neo4jMatchers.hasProperty;
import static org.neo4j.graphdb.Neo4jMatchers.inTx;
import static org.neo4j.helpers.collection.MapUtil.map;

public class CypherUpdateMapTest
{
    private ExecutionEngine engine;
    private GraphDatabaseCypherService gdb;
    private QuerySession session = QueryEngineProvider.embeddedSession();

    @Test
    public void updateNodeByMapParameter() throws Throwable
    {
        engine.executeQuery(
                "CREATE (n:Reference) SET n = {data} RETURN n" ,
                map( "data",
                        map("key1", "value1", "key2", 1234)
                ),
                session
        );

        Node node1 = getNodeByIdInTx( 0 );

        assertThat( node1, inTxS( hasProperty( "key1" ).withValue( "value1" ) ) );
        assertThat( node1, inTxS( hasProperty( "key2" ).withValue( 1234 ) ) );

        engine.executeQuery(
                "MATCH (n:Reference) SET n = {data} RETURN n",
                map( "data",
                        map("key1", null, "key3", 5678)
                ),
                session
        );

        Node node2 = getNodeByIdInTx( 0 );

        assertThat( node2, inTxS( not( hasProperty( "key1" ) ) ) );
        assertThat( node2, inTxS( not( hasProperty( "key2" ) ) ) );
        assertThat( node2, inTxS( hasProperty( "key3" ).withValue(5678) ) );
    }

    public <T> Matcher<? super T> inTxS( final Matcher<T> inner )
    {
        return inTx( gdb.getGraphDatabaseService(), inner, false );
    }

    private Node getNodeByIdInTx( int nodeId )
    {
        try ( Transaction ignored = gdb.beginTransaction(KernelTransaction.Type.explicit, AccessMode.READ) )
        {
            return gdb.getNodeById( nodeId );
        }
    }

    @Before
    public void setup()
    {
        gdb = new GraphDatabaseCypherService( new TestGraphDatabaseFactory().newImpermanentDatabase() );
        engine = new ExecutionEngine( gdb, NullLogProvider.getInstance() );
    }

    @After
    public void cleanup()
    {
        gdb.getGraphDatabaseService().shutdown();
    }
}
